/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* todo:
   write postfix passer
*/

#define CORE_PRIVATE
#include "httpd.h"
#include "http_protocol.h"
#include "http_config.h"
#include "http_connection.h"
#include "http_core.h"
#include "http_request.h"
#include "http_main.h"
#include "http_log.h"
#include "ap_config.h"
#include "ap_mmn.h"
#include "apr_lib.h"
#include "util_script.h"
#include "apr_buckets.h"
#include "scoreboard.h"
#include "apr_strings.h"
#include "apr_pools.h"
#include "apr_network_io.h"

#include "mod_smtpd.h"
#include "smtp.h"

module AP_MODULE_DECLARE_DATA smtpd_module;

static apr_hash_t *smtpd_handlers;

/* Implement 'smtpd_run_unrecognized_command'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode,
				    unrecognized_command,
				    (smtpd_conn_rec *scr, smtpd_return_data *in,
				     char *command, char *data),
				    (scr, in, command, data),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_connect'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode, connect,
				    (smtpd_conn_rec *scr, smtpd_return_data *in),
				    (scr, in),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_reset_transaction'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode,
				    reset_transaction,
				    (smtpd_conn_rec *scr),
				    (scr),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_helo'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode, helo,
				    (smtpd_conn_rec *scr, smtpd_return_data *in,
				     char *str),
				    (scr, in, str),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_ehlo'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode, ehlo,
				    (smtpd_conn_rec *scr, smtpd_return_data *in,
				     char *str),
				    (scr, in, str),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_mail'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode, mail,
				    (smtpd_conn_rec *scr, smtpd_return_data *in,
				     char *str),
				    (scr, in, str),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_rcpt'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode, rcpt,
				    (smtpd_conn_rec *scr, smtpd_return_data *in,
				     char *str),
				    (scr, in, str),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_vrfy'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode, vrfy,
				    (smtpd_conn_rec *scr, smtpd_return_data *in,
				     char *str),
				    (scr, in, str),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_quit'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode, quit,
				    (smtpd_conn_rec *scr, smtpd_return_data *in),
				    (scr, in),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_data'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode, data,
				    (smtpd_conn_rec *scr, smtpd_return_data *in),
				    (scr, in),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_data_post'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode,
				    data_post,
				    (smtpd_conn_rec *scr, smtpd_return_data *in),
				    (scr, in),
				    SMTPD_DECLINED, SMTPD_DECLINED);
/* Implement 'smtpd_run_queue'. */
APR_IMPLEMENT_OPTIONAL_HOOK_RUN_ALL(smtpd, SMTPD, smtpd_retcode, queue,
				    (smtpd_conn_rec *scr, smtpd_return_data *in),
				    (scr, in),
				    SMTPD_DECLINED, SMTPD_DECLINED);
				      
/* public methods */
/* functions other modules can use */

/* should be called at smtpd_hook_connect */
SMTPD_DECLARE_NONSTD(void)
smtpd_register_extension(smtpd_conn_rec *scr, const char *line)
{
  (*((char **)apr_array_push(scr->extensions))) =
    apr_pstrdup(scr->p, line);
}

/* how to reset the transaction */
SMTPD_DECLARE_NONSTD(void)
smtpd_reset_transaction(smtpd_conn_rec *scr) {
  /* REVIEW: don't know whether to run clear request first
   * then run reset hooks, or run reset hooks then clear request
   * depends on whether hooks want to save info before it gets cleared out
   * or if they want to overwrite default values in the request rec
   */
  smtpd_run_reset_transaction(scr);

  smtpd_clear_trans_rec(scr->transaction);
}

/* friend methods */
/* only our sources can call these */

apr_status_t
smtpd_getline(smtpd_conn_rec *scr, char *data, apr_size_t dlen)
{
  apr_status_t rc;
  apr_bucket *e;
  const char *str;
  char *pos, *last_char = data;
  apr_size_t len, bytes_handled = 0;

  while (1) {
    rc = ap_get_brigade(scr->c->input_filters, scr->bb_in, AP_MODE_GETLINE,
			APR_BLOCK_READ, 0);
    if (rc != APR_SUCCESS) return rc;
    
    while(!APR_BRIGADE_EMPTY(scr->bb_in)) {
      e = APR_BRIGADE_FIRST(scr->bb_in);
      
      rc = apr_bucket_read(e, &str, &len, APR_BLOCK_READ);
      if (rc != APR_SUCCESS) return rc;      
      apr_bucket_delete(e);
           
      if (len == 0) continue;
 
      /* Would this overrun our buffer?  If so, we'll die. */
      if (dlen < bytes_handled + len) {
	if (data) {
	  /* ensure this string is NUL terminated */
	  if (bytes_handled > 0) {
	    data[bytes_handled-1] = '\0';
	  }
	  else {
	    data[0] = '\0';
	  }
	}
	return APR_ENOSPC;
      }

      /* Just copy the rest of the data to the end of the old buffer. */
      pos = data + bytes_handled;
      memcpy(pos, str, len);
      last_char = pos + len - 1;
            
      /* We've now processed that new data - update accordingly. */
      bytes_handled += len;
    }
    
    /* brigade was empty from the beginning, no block? */
    if (!bytes_handled)
      return APR_EGENERAL;

    /* If we got a full line of input, stop reading */
    if (last_char && (*last_char == APR_ASCII_LF)) {
      break;
    }
  }

  /* Now NUL-terminate the string at the end of the line; 
   * if the last-but-one character is a CR, terminate there */
  if (last_char > data && last_char[-1] == APR_ASCII_CR) {
    last_char--;
  }

  *last_char = '\0';
  return APR_SUCCESS;
}

apr_status_t
smtpd_respond_multiline(smtpd_conn_rec *scr, int code, char *messages[])
{
  char *str;
  int i;

  for (i = 0; messages[i]; i++) {
    str = messages[i+1] ? "%d %s\r\n" : "%d-%s\r\n";
    ap_fprintf(scr->c->output_filters, scr->bb_out, str, code, messages[i]);
    ap_fflush(scr->c->output_filters, scr->bb_out);
  }

  return APR_SUCCESS;
}

apr_status_t
smtpd_respond_oneline(smtpd_conn_rec *scr, int code, char *message)
{
  ap_fprintf(scr->c->output_filters, scr->bb_out, "%d %s\r\n", code, message);
  ap_fflush(scr->c->output_filters, scr->bb_out);

  return APR_SUCCESS;
}

void
smtpd_clear_trans_rec(smtpd_trans_rec *str) {
  apr_pool_clear(str->p);
  str->state_vector = SMTPD_STATE_GOT_NOTHING;
  str->tfp = NULL;
  str->extended = SMTPD_PROTOCOL_SMTP;
  str->rcpt_to = apr_array_make(str->p, 5, sizeof(char *));
  str->mail_from = NULL;
  str->helo = NULL;
}

apr_hash_t *
smtpd_get_handlers() {
  return smtpd_handlers;
}

/* private methods */
/* only used in this file */

/* can overwrite currently registered handlers */
static void
smtpd_register_handler(char *key, smtpd_handler *func, const char *help_text,
		       void *data, apr_pool_t *p)
{
  char *dupkey = apr_pstrdup(p, key);
  smtpd_handler_st *hand = apr_palloc(p, sizeof(*hand));

  hand->func = func;
  hand->data = data;
  hand->help_text = help_text;

  ap_str_tolower(dupkey);
  apr_hash_set(smtpd_handlers, dupkey, APR_HASH_KEY_STRING, hand);
}

/* Creates the main request record for the connection */
static smtpd_conn_rec *
smtpd_create_conn_rec(conn_rec *conn)
{
  apr_pool_t *p;
  smtpd_conn_rec *scr;
  apr_pool_t *sp;
  smtpd_trans_rec *str;

  scr = apr_pcalloc(conn->pool, sizeof(*scr));
  apr_pool_create(&p, conn->pool);
  scr->p            = p;
  scr->c            = conn;
  scr->s            = conn->base_server;

  scr->extensions = apr_array_make(scr->p, 5, sizeof(char *));

  scr->bb_in = apr_brigade_create(scr->p, scr->c->bucket_alloc);
  scr->bb_out = apr_brigade_create(scr->p, scr->c->bucket_alloc);

  //  r->request_config  = ap_create_request_config(r->pool);
  
  apr_socket_t *csd =
    ((core_net_rec *)conn->input_filters->ctx)->client_socket;
  apr_socket_timeout_set(csd, APR_INT64_C(10000000000000));

  /* create transaction rec */
  str = apr_pcalloc(scr->p, sizeof(*str));

  apr_pool_create(&sp, scr->p);
  str->p = sp;
  smtpd_clear_trans_rec(str);

  scr->transaction = str;

  return scr;
}

/* process connection hook */
static int
process_smtp_connection(conn_rec *c)
{
  smtpd_conn_rec *scr;
  smtpd_svr_config_rec *pConfig =
    ap_get_module_config(c->base_server->module_config,
			 &smtpd_module);

  if (!pConfig->bEnabled) {
    return DECLINED;
  }

  scr = smtpd_create_conn_rec(c);
  //  ap_update_child_status(scr->c->sbh, SERVER_BUSY_KEEPALIVE, r);

  smtpd_process_connection_internal(scr);

  return OK;
}

/* creates server config  */
static void *
smtpd_create_server_config(apr_pool_t *p, server_rec *s)
{
  smtpd_svr_config_rec *pConfig = apr_pcalloc(p, sizeof(*pConfig));

  pConfig->bEnabled = 0;
  pConfig->sId = apr_pcalloc(p, 512);
  apr_cpystrn(pConfig->sId, "mod_smtpd", 512);
  pConfig->max_data = 0x80000;
  return pConfig;
}

/* sets protocol status in server config */
static const char *
set_protocol_status(cmd_parms *cmd, 
		    void *struct_ptr, 
		    int arg)
{
  smtpd_svr_config_rec *pConfig =
    ap_get_module_config(cmd->server->module_config,
			 &smtpd_module);
  pConfig->bEnabled = arg ? 1 : 0;

  return NULL;
}

/* sets server id string in server config */
static const char *
set_id_string(cmd_parms *cmd,
	      void *struct_ptr,
	      const char *arg)
{
  smtpd_svr_config_rec *pConfig =
    ap_get_module_config(cmd->server->module_config,
			 &smtpd_module);

  apr_cpystrn(pConfig->sId, arg, 512);
    
  return NULL;
}

/* sets server id string in server config */
static const char *
set_max_data_size(cmd_parms *cmd,
		  void *struct_ptr,
		  const char *arg)
{
  smtpd_svr_config_rec *pConfig =
    ap_get_module_config(cmd->server->module_config,
			 &smtpd_module);

  pConfig->max_data = atoi(arg);
    
  return NULL;
}

static const command_rec smtpd_cmds[] = {
  AP_INIT_FLAG("SmtpProtocol", set_protocol_status, NULL,  RSRC_CONF,
	       "Whether this server is serving the SMTP protocol."
	       "Default: Off"),
  AP_INIT_TAKE1("SmtpServerId", set_id_string, NULL,  RSRC_CONF,
		"Server identification to advertise."
		"Default: mod_smtpd"),
  AP_INIT_TAKE1("SmtpMaxDataSize", set_max_data_size, NULL,  RSRC_CONF,
		"Maximum input to receive after data command in bytes."
		"Default: 8388608"),

  { NULL }
};

/* registers httpd hooks */
static void
register_hooks (apr_pool_t *p)
{
  /* register connection processor  */
  ap_hook_process_connection(process_smtp_connection, NULL, 
			     NULL, APR_HOOK_MIDDLE);

  smtpd_handlers = apr_hash_make(p);

  smtpd_register_handler("EHLO", HANDLER_FUNC(ehlo), "ehlo", NULL, p);
  smtpd_register_handler("HELO", HANDLER_FUNC(helo), "helo", NULL, p);
  smtpd_register_handler("MAIL", HANDLER_FUNC(mail), "mail", NULL, p);
  smtpd_register_handler("RCPT", HANDLER_FUNC(rcpt), "rcpt", NULL, p);
  smtpd_register_handler("DATA", HANDLER_FUNC(data), "data", NULL, p);
  smtpd_register_handler("RSET", HANDLER_FUNC(rset), "rset", NULL, p);
  smtpd_register_handler("NOOP", HANDLER_FUNC(noop), "noop", NULL, p);
  smtpd_register_handler("QUIT", HANDLER_FUNC(quit), "quit", NULL, p);
  smtpd_register_handler("VRFY", HANDLER_FUNC(vrfy), "vrfy", NULL, p);
}

module AP_MODULE_DECLARE_DATA smtpd_module = {
  STANDARD20_MODULE_STUFF,
  NULL,                       /* create per-directory config structure */
  NULL,                       /* merge per-directory config structures */
  smtpd_create_server_config, /* create per-server config structure  */
  NULL,                       /* merge per-server config structures */
  smtpd_cmds,                 /* command apr_table_t */
  register_hooks              /* register hooks */
};
