/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef SMTP_H
#define SMTP_H

#include "mod_smtpd.h"

#ifdef __cplusplus
extern "C" {
#endif

  /* SMTP handlers registration */
#define HANDLER_PROTOTYPE smtpd_conn_rec *scr, char *buffer, smtpd_return_data *in_data, void *data

#define HANDLER_FUNC(name)  smtpd_handler_##name
#define HANDLER_DECLARE(name) SMTPD_DECLARE(int) HANDLER_FUNC(name) (HANDLER_PROTOTYPE)

  typedef SMTPD_DECLARE(int) smtpd_handler(HANDLER_PROTOTYPE);

  typedef struct {
    int bEnabled; /* Is SMTP Enabled? */
    char *sId; /* Server ID String */
    apr_size_t max_data;
  } smtpd_svr_config_rec;

  typedef struct {
    smtpd_handler *func;
    const char *help_text;
    void *data;
  } smtpd_handler_st;

  void
  smtpd_process_connection_internal(smtpd_conn_rec *str);

  void
  smtpd_clear_trans_rec(smtpd_trans_rec *);

  apr_hash_t *
  smtpd_get_handlers(void);

  apr_status_t
  smtpd_getline(smtpd_conn_rec *scr, char *data, apr_size_t dlen);

  apr_status_t
  smtpd_respond_multiline(smtpd_conn_rec *scr, int code, char *messages[]);

  apr_status_t
  smtpd_respond_oneline(smtpd_conn_rec *scr, int code, char *message);
    
  HANDLER_DECLARE(ehlo);
  HANDLER_DECLARE(helo);
  HANDLER_DECLARE(mail);
  HANDLER_DECLARE(rcpt);
  HANDLER_DECLARE(data);
  HANDLER_DECLARE(rset);
  HANDLER_DECLARE(noop);
  HANDLER_DECLARE(quit);
  HANDLER_DECLARE(vrfy);

#ifdef __cplusplus
}
#endif

#endif
